/***************************************************************************************
 *
 * Intel(R) Trace Collector Library.
 *
 * Copyright(C) 1997-2015 Intel Corporation.
 * All Rights Reserved.
 * 
 * The source code contained or described herein and all documents
 * related to the source code ("Material") are owned by Intel
 * Corporation or its suppliers or licensors. Title to the Material
 * remains with Intel Corporation or its suppliers and licensors. The
 * Material contains trade secrets and proprietary and confidential
 * information of Intel or its suppliers and licensors. The Material
 * is protected by worldwide copyright and trade secret laws and
 * treaty provisions. No part of the Material may be used, copied,
 * reproduced, modified, published, uploaded, posted, transmitted,
 * distributed, or disclosed in any way without Intel's prior express
 * written permission.
 * 
 * No license under any patent, copyright, trade secret or other
 * intellectual property right is granted to or conferred upon you by
 * disclosure or delivery of the Materials, either expressly, by
 * implication, inducement, estoppel or otherwise. Any license under
 * such intellectual property rights must be express and approved by
 * Intel in writing.
 * 
 ***************************************************************************************/

/**
 * @name API
 */
/*@{*/

/**
 *
 * @file
 *
 * This is a documentation of all current Intel(R) Trace Collector API functions.
 * They use the VT prefix because this product was formerly called Vampirtrace.
 *
 * @version 20130918
 * @author Intel Corporation
 * @par
 *
 *
 * This version of the ITC library is thread-safe in the sense that all of
 * their API functions can be called by several threads at the same time.
 * Some API functions can really be executed concurrently, others protect
 * global data with POSIX mutices.
 *
 * In this documentation the term "threads" refers to one stream of events
 * and a "process" is a collection of threads. Processes are enumerated
 * in the same way as MPI enumerates them in MPI_COMM_WORLD (0 till n-1).
 * Each process may have a different number of threads, which are
 * enumerated within their process. So each thread is exactly identified
 * by its thread index and its process index. If multithreading is not
 * supported, then all thread indices are 0 and a process is the same as
 * a thread.
 *
 * A thread can enter and leave states and log events during its execution.
 * A state is identified by two strings: the "activity" groups several
 * "symbols" together, f.i. MPI contains MPI_Init. This is sometimes written
 * as MPI:MPI_Init. Two different states may have the same symbol name, as
 * long as their activities are different. A "class" is a generalization of
 * this hierarchy concept and can be used instead of an activity in a state,
 * but also for counters.
 *
 * ITC keeps a stack of states for each thread: entering a state adds one entry
 * and leaving it removes it. This stack does not have to be empty when
 * VT_finalize()/MPI_Finalize() are called or the thread terminates.
 * ITC will automatically generate leave records for all states that are
 * still on the stack.
 *
 * Many API functions refer to strings or entities like states with a handle.
 * In previous versions of ITC this handle had to be chosen carefully by the
 * application so that the handle was unique and consistent across all
 * processes. Such a handle must be a number >= 0 and in the case of symbols
 * also < VT_MAX_USERCODE.
 *
 * This functionality is still supported, but an easier method is having
 * ITC assign a handle itself. These handles are stored in C or Fortran ints.
 * They can be copied, but must not be casted or modified by the application.
 * Allocating such a handle for a string or entity that has been assigned a
 * handle already is not an error. The second call will simple return the
 * existing handle.
 *
 * Functions that get a handle as input always accept both old and new handles.
 *
 * ITC API functions that get strings as input do not modify these strings. They copy
 * them, so the buffer in the application can be reused after the API function
 * has finished. Strings can have arbitrary lengths.
 *
 * In general the API functions are also provided as Fortran subroutines which
 * take exactly the same parameters as the C versions, but get one more integer
 * parameter where they store the error code. Their names however might differ slightly
 * and some are functions that return a value just like in C, so for each function the
 * corresponding Fortran interface is given.
 *
 **/
/***************************************************************************************/

#ifndef _VT_H_
#define _VT_H_

#if defined(__cplusplus)
/* tell C++ compilers that C functions follow */
extern          "C" {
#endif

/**
 * API version constant.
 *
 * It is incremented each time the API changes, even if the change
 * does not break compatibility with the existing API. It can be used
 * to determine at compile time how to call the API, like this:
 *
\verbatim 
   #if VT_VERSION > 4000
      do something
   #else
      do something different
   #endif
\endverbatim
 *
 * VT_version() provides the same information at runtime.
 *
 * To check whether the current revision of the API is still compatible
 * with the revision of the API that the application was written against,
 * compare against both VT_VERSION and VT_VERSION_COMPATIBILITY, as
 * shown below.
 */
#define VT_VERSION                7000

/**
 * Oldest API definition which is still compatible with the
 * current one.
 *
 * This is set to the current version each time an API change can
 * break programs written for the previous API. For example, a program
 * written for VT_VERSION 2090 will work with API 3000 if
 * VT_VERSION_COMPATIBILITY remained at 2090. It may even work without
 * modifications when VT_VERSION_COMPATIBILITY was increased to
 * 3000, but this cannot be determined automatically and will require
 * a source code review.
 *
 * Here is a usage example:
\verbatim
   #define APP_VT_VERSION 1000 // API version used by APP
   #ifdef VT_VERSION_COMPATIBILITY > APP_EXPECTED_VT_VERSION
   # error "VT.h is no longer compatible with APP's usage of the API"
   #endif
   #ifdef VT_VERSION < APP_EXPECTED_VT_VERSION
   # error "VT.h is not recent enough for APP"
   #endif
\endverbatim
 */
#define VT_VERSION_COMPATIBILITY  2090


/** Max user-defined symbol code */
#define VT_MAX_USERCODE	 32768

/** special value for VT_end(): let ITC choose the correct state */
#define VT_NOSTATE       -1

/** special value for VT_funcdef(): put function into the default class "Application" */
#define VT_NOCLASS       0x00000000

/** special SCL handle: no location available */
#define VT_NOSCL        -0x7FFFFFFD

/** error codes returned by ITC API */
typedef enum _VT_ErrorCode {
    VT_OK = 0,                       /**< OK. */
    VT_ERR_NOLICENSE      = -1,      /**< no valid license found. */
    VT_ERR_NOTIMPLEMENTED = -2,      /**< Not (yet?) implemented. */
    VT_ERR_NOTINITIALIZED = -3,      /**< Not initialised. */
    VT_ERR_BADREQUEST     = -4,      /**< Invalid request type */
    VT_ERR_BADSYMBOLID    = -5,      /**< Wrong symbol id. */
    VT_ERR_BADSCLID       = -6,      /**< wrong SCL id. */
    VT_ERR_BADSCL         = -7,      /**< wrong SCL. */
    VT_ERR_BADFORMAT      = -8,      /**< wrong format. */
    VT_ERR_BADKIND        = -9,      /**< Wrong kind found. */
    VT_ERR_NOMEMORY       = -10,     /**< Could not get memory. */
    VT_ERR_BADFILE        = -11,     /**< Error while handling file. */
    VT_ERR_FLUSH          = -12,     /**< Error while flushing. */
    VT_ERR_BADARG         = -13,     /**< wrong argument. */
    VT_ERR_NOTHREADS      = -14,     /**< no worker threads. */
    VT_ERR_BADINDEX       = -15,     /**< wrong thread index. */
    VT_ERR_COMM           = -16,     /**< communication error. */  
    VT_ERR_INVT           = -17,     /**< ITC API called while inside an ITC function */
    VT_ERR_IGNORE         = -18      /**< non-fatal error code */
} VT_ErrorCode_t;

/**
 * Logging send/receive events evaluates process rank local within
 * the active communicator, and matches events only if they
 * are taking place in the same communicator (in other words,
 * it is the same behaviour as in MPI).
 *
 * Defining new communicators is currently not supported,
 * but the predefined ones can be used.
 */
typedef enum _VT_CommIDs {
    VT_COMM_INVALID = 0,     /**< invalid ID, do not pass to ITC */
    VT_COMM_WORLD = 1,       /**< global ranks are the same as local ones */
    VT_COMM_SELF = 2         /**< communicator that only contains the active process */
} VT_CommIDs_t;

/**
 * These are operation ids that can be passed to VT_log_op().
 * Their representation in the trace file matches that of the
 * equivalent MPI operation.
 *
 * User-defined operations are currently not supported.
 */
typedef enum _VT_OpTypes {
    VT_OP_INVALID        = 0,       /**< undefined operation, should not be passed to ITC */
    VT_OP_BARRIER,
    VT_OP_IBARRIER,
    VT_OP_BCAST,
    VT_OP_IBCAST,
    VT_OP_GATHER,
    VT_OP_IGATHER,
    VT_OP_GATHERV,
    VT_OP_IGATHERV,
    VT_OP_SCATTER,
    VT_OP_ISCATTER,
    VT_OP_SCATTERV,
    VT_OP_ISCATTERV,
    VT_OP_ALLGATHER,
    VT_OP_IALLGATHER,
    VT_OP_ALLGATHERV,
    VT_OP_IALLGATHERV,
    VT_OP_ALLTOALL,
    VT_OP_IALLTOALL,
    VT_OP_ALLTOALLV,
    VT_OP_IALLTOALLV,
    VT_OP_ALLTOALLW,
    VT_OP_IALLTOALLW,
    VT_OP_NEIGHBOR_ALLGATHER,
    VT_OP_INEIGHBOR_ALLGATHER,
    VT_OP_NEIGHBOR_ALLGATHERV,
    VT_OP_INEIGHBOR_ALLGATHERV,
    VT_OP_NEIGHBOR_ALLTOALL,
    VT_OP_INEIGHBOR_ALLTOALL,
    VT_OP_NEIGHBOR_ALLTOALLV,
    VT_OP_INEIGHBOR_ALLTOALLV,
    VT_OP_NEIGHBOR_ALLTOALLW,
    VT_OP_INEIGHBOR_ALLTOALLW,
    VT_OP_REDUCE,
    VT_OP_REDUCE_LOCAL,
    VT_OP_IREDUCE,
    VT_OP_ALLREDUCE,
    VT_OP_IALLREDUCE,
    VT_OP_REDUCE_SCATTER,
    VT_OP_REDUCE_SCATTER_BLOCK,
    VT_OP_IREDUCE_SCATTER,
    VT_OP_IREDUCE_SCATTER_BLOCK,
    VT_OP_SCAN,
    VT_OP_ISCAN,
    VT_OP_EXSCAN,
    VT_OP_IEXSCAN,
    VT_OP_COUNT   /**< number of predefined operations */
} VT_OpTypes_t;



/** @name ITC main API */
/*@{*/

/** 
 * Initialize ITC and underlying communication.
 *
 * VT_initialize(), VT_getrank(), VT_finalize() can be used to write applications or
 * libraries which work both with and without MPI, depending on whether they are linked
 * with libVT.a plus MPI or with libVTcs.a (distributed tracing) and no MPI.
 *
 * If the MPI that ITC was compiled for provides MPI_Init_thread(), then
 * VT_init() will call MPI_Init_thread() with the parameter \a required set
 * to MPI_THREAD_FUNNELED. This is sufficient to initialize multithreaded
 * applications where only the main thread calls MPI. If your application
 * requires a higher thread level, then either use MPI_Init_thread()
 * instead of VT_init() or (if VT_init() is called e.g. by your runtime
 * environment) set the environment variable VT_THREAD_LEVEL to a value of
 * 0 till 3 to choose thread levels MPI_THREAD_SINGLE till MPI_THREAD_MULTIPLE.
 *
 * It is not an error to call VT_initialize() twice or after a MPI_Init().
 *
 * In a MPI application written in C the program's parameters must be
 * passed, because the underlying MPI might require them. Otherwise they
 * are optional and 0 resp. a NULL pointer may be used. If parameters are
 * passed, then the number of parameters and the array itself may be
 * modified, either by MPI or ITC itself.
 *
 * ITC assumes that (*argv)[0] is the executable's name and uses this
 * string to find the executable and as the basename for the default
 * logfile name. Other parameters are ignored unless there is the
 * special "--tracecollector-args" parameters: then all following parameters
 * are interpreted as configuration options, written with a double
 * hyphen as prefix and a hyphen instead of underscores (e.g.
 * --tracecollector-args --logfile-format BINARY --logfile-prefix /tmp).
 * These parameters are then removed from the argv array, but not freed.
 * To continue with the program's normal parameters, --tracecollector-args-end
 * may be used. There may be more than one block of ITC arguments on the
 * command line.
 *
 * @par Fortran
 *  VTINIT( ierr )
 *
 * @par Java
 *  void initialize( )
 *
 * @param    argc      a pointer to the number of command line arguments
 * @param    argv      a pointer to the program's command line arguments
 *
 * @return error code
 **/
extern int  VT_initialize( int * argc, char *** argv );

/**
 * Finalize ITC and underlying communication. 
 *
 * It is not an error to call VT_finalize() twice or after a MPI_Finalize().
 *
 * @par Fortran
 *  VTFINI( ierr )
 *
 * @par Java
 *  void fini( )
 *
 * @return error code
 **/
extern int  VT_finalize( void );

/** a callback without arguments or return code */
typedef void (*VT_Callback_t)( void );

/**
 * Sets a callback which is called by ITC at the beginning of
 * finalization. This function may use the ITC API to log
 * events.
 *
 * Only one callback can be stored per process, setting another
 * or NULL removes the previous callback.
 *
 * @param callback   pointer to the callback 
 *
 * @return error code
 **/
extern int  VT_setfinalizecallback( VT_Callback_t callback );

typedef enum _VT_DescriptionType {
    VT_DESCRIPTION_LIB,       /**< library base name, e.g. libVT */
    VT_DESCRIPTION_PURPOSE,   /**< main purpose of the library, e.g. "MPI tracing" */
    VT_DESCRIPTION_DATE,      /**< date when lib was compiled */
    VT_DESCRIPTION_PLATFORM,  /**< platform ID string */
    VT_DESCRIPTION_VERSION,   /**< version string */
    VT_DESCRIPTION_API,       /**< a list of all public functions in the library, separated by \n */
    VT_DESCRIPTION_MAX        /**< end of enumeration */
} VT_DescriptionType_t;

/**
 * Returns a pointer that describes certain aspects of the library
 * that implements the ITC API. This call can be used by code that
 * is compatible with any library implementing the API, but
 * nevertheless wants to identify the implementation.
 *
 * @param type     what kind of information is requested (VT_DESCRIPTION_LIB, ... )
 * @return a short string or NULL of the requested information is unavailable
 */
extern const char *VT_getdescription( int type );

/**
 * Provides information about the API revision implemented by the
 * library. Can be called at any time, i.e. ITC does not have to
 * be initialized, and will never fail. Passing a null pointer as
 * address of return values when not interested in the value is
 * valid.
 *
 * @par Fortran
 *  VTVERSION( appversion, vtversion, backversion, compatible )
 *
 * @par Java
 *  Only supports version check, returns compatibility:
 *  void boolean version( appversion )
 *
 * @param  appversion      version of the API expected by the caller,
 *                         similar to APP_VT_VERSION in the
 *                         VT_VERSION_COMPATIBILITY example; pass 0 if
 *                         not used
 * @retval vtversion       current revision of the API, see VT_VERSION
 * @retval backversion     oldest API revision still guaranteed to work,
 *                         see VT_VERSION_COMPATIBILITY
 * @retval compatible      1 if ITC is sure that the implementation is
 *                         compatible with appversion,
 *                         0 otherwise
 */
extern void VT_version( int appversion,
                        int *vtversion,
                        int *backversion,
                        int *compatible );

/**
 * Returns non-zero value if implementation of the ITC API is
 * available. Zero is returned if only dummy implementations
 * are provided: calling them will not have an effect and can
 * be avoided to improve performance.
 *
 * @Fortran
 *   INTEGER VTENABLED()
 */
extern int VT_enabled( void );

/** 
 * A dummy implementation of the VT_dynamic.h call VT_loadlibrary().
 *
 * By having it defined also in VT.h you can write code which
 * links either statically against ITC or loads it dynamically.
 * To switch between the two modes, either include VT.h or
 * VT_dynamic.h.
 */
#define VT_loadlibrary( lib ) 0

/**
 * A callback which is called by ITC once at the end of initialization.
 * The application may provide its own variable with this name (to
 * set the pointer at link time) or set this variable before
 * initializing ITC. When the callback is called, it may use all
 * ITC API calls.
 */
extern VT_Callback_t VT_APP_initcallback;

/**
 * Get process index (same as MPI rank within MPI_COMM_WORLD).
 * Beware that this number is not unique in applications with
 * dynamic process spawning.
 *
 * @par Fortran
 *  VTGETRANK( rank, ierr )
 *
 * @par Java
 *  int getRank()
 *
 * @retval rank process index is stored here
 * @return error code
 **/
extern int  VT_getrank( int *rank );

/**
 * Get process rank inside the given communicator.
 * This is useful if the application uses dynamic process
 * spawning.
 *
 * @par Fortran
 *  VTCOMMRANK( commid, rank, ierr )
 *
 * @par Java
 *  int commRank( commid )
 *
 * @param  commid  the communicator number (e.g. VT_COMM_WORLD)
 * @retval rank    process rank is stored here
 * @return error code
 **/
extern int  VT_comm_rank( int commid, int *rank );

/**
 * Get number of processes in the given communicator.
 * This is useful if the application uses dynamic process
 * spawning.
 *
 * @par Fortran
 *  VTCOMMRANK( commid, rank, ierr )
 *
 * @par Java
 *  int commSize( commid )
 *
 * @param  commid  the communicator number (e.g. VT_COMM_WORLD)
 * @retval size    number of processes is stored here
 * @return error code
 **/
extern int  VT_comm_size( int commid, int *size );


/**
 * Returns an opaque time stamp, or VT_ERR_NOTINITIALIZED. In contrast
 * to previous versions this time stamp no longer represents
 * seconds. Use VT_timeofday() for that instead. The result of
 * VT_timestamp() can be copied verbatim and given to other API calls,
 * but nothing else.
 *
 * @par Fortran
 *  DOUBLE PRECISION VTSTAMP( )
 *
 * @par Java
 *  double timeStamp( )
 *
 **/
extern double   VT_timestamp( void );

/**
 * Returns the systems wall clock time in seconds, or VT_ERR_NOTINITIALIZED.
 *
 * @par Fortran
 *  DOUBLE PRECISION VTTIMEOFDAY( )
 *
 * @par Java
 *  double timeOfDay( )
 *
 **/
extern double   VT_timeofday( void );

/**
 * Returns point in time in seconds when process started, or VT_ERR_NOTINITIALIZED.
 *
 * @par Fortran
 *  DOUBLE PRECISION VTTIMESTART( )
 *
 * @par Java
 *  double timeStart( )
 * 
 */
extern double   VT_timestart( void );

/**
 * Gathers data needed for clock synchronization.
 * This is a collective call, so all processes which
 * were started together must call this function or
 * it will block.
 *
 * This function does not work if processes were
 * spawned dynamically.
 *
 * @par Fortran
 *  VTTIMESYNC( ierr )
 *
 * @par Java
 *  void timeSync( )
 *
 */
extern int      VT_timesync( void );

/**
 * Flushes all trace records from memory into the flush file.
 * The location of the flush file is controlled by options in the config file.
 * Flushing will be recorded in the trace file as entering and leaving the state
 * VT_API:TRACE_FLUSH with time stamps that indicate the duration of the flushing.
 * Automatic flushing is recorded as VT_API:AUTO_FLUSH.
 *
 * @par Fortran
 *  VTFLUSH( ierr )
 *
 * @par Java
 *  void flush( )
 *
 * @return error code
 */
extern int  VT_flush( void );

/**
 * Defines the numeric \a statehandle as shortcut for a state.
 * This function will become obsolete and should not be used
 * for new code. Both \a symname and \a activity may consist
 * of more than one component, separated by a colon (:).
 *
 * Leading and trailing colons are ignored. Several colons in
 * a row are treated as just one separator.
 *
 * @par Fortran
 *  VTSYMDEF( code, symname, activity, ierr )
 *
 * @param statehandle numeric value chosen by the application
 * @param symname name of the symbol
 * @param activity name of activity this symbol belongs to
 * @return error code
 */
extern int  VT_symdef( int statehandle, const char * symname, const char * activity );

/**
 * Obsolete version of VT_symdef(). The additional parameter is ignored.
 */
extern int  VT_symdefl( int statehandle, const char * symname, const char * activity, int sclhandle );

/**
 * Allocates a handle for a class name.
 * The \a classname may consist of several components separated by a colon (:).
 * Leading and trailing colons are ignored. Several colons in
 * a row are treated as just one separator.
 *
 *
 * @par Fortran
 *  VTCLASSDEF( classname, classhandle, ierr )
 *
 * @par Java
 *  int classDef( final String classname )
 *
 * @param classname      name of the class
 * @retval classhandle   the int it points to is set by ITC
 * @return error code
 */
extern int  VT_classdef( const char * classname, int * classhandle );

/**
 * Allocates a handle for a state.
 *
 * The \a symname may consist of several components separated by a colon (:).
 * If that's the case, then these become the parent class(es).
 * Leading and trailing colons are ignored. Several colons in
 * a row are treated as just one separator.
 *
 * This is a replacement for VT_symdef() which doesn't require the
 * application to provide a unique numeric handle.
 *
 * @par Fortran
 *  VTFUNCDEF( symname, classhandle, statehandle, ierr )
 *
 * @par Java
 *  int funcDef( final String symname, int classhandle )
 *
 * @param symname      name of the symbol
 * @param classhandle  handle for the class this symbol belongs to, created with VT_classdef(),
 *                     or VT_NOCLASS, which is an alias for "Application" if the symname doesn't
 *                     contain a class name and ignored otherwise
 * @retval statehandle the int it points to is set by ITC
 * @return error code
 */
extern int  VT_funcdef( const char * symname, int classhandle, int * statehandle );

/**
 * Allocates a handle for a source code location (SCL).
 *
 * @par Fortran
 *  VTSCLDEF( file, line_nr, sclhandle, ierr )
 *
 * @par Java
 *  int sclDef( final String file, int line_nr )
 * 
 * @param file file name
 * @param line_nr line number in this file, counting from 1
 * @retval sclhandle the int it points to is set by ITC
 * @return error code
 */
extern int  VT_scldef( const char * file, int line_nr, int * sclhandle );

/**
 * Allocates a handle for a source code location (SCL) handle which
 * refers to the current call stack.
 *
 * This SCL can then be used in several API calls without having to
 * repeat the stack unwinding each time. Which stack frames are
 * preserved and which are skipped is determined by the PCTRACE
 * configuration option, but can be overridden with function
 * parameters.
 *
 * Special support is available for recording source code locations
 * from inside signal handlers by calling this function with the pc
 * and stackframe parameters different from NULL. Other usages of
 * these special parameters include:
 * - remembering the stack frame in those API calls of a library
 *   that are invoked directly by the application, then at arbitrary
 *   points in the library do stack unwinding based on that stack
 *   frame to catch just the application code
 * - defining a source code location ID for a specific program
 *   counter value
 *
 * Here is an example of the usage of this call inside a library which
 * implements a message send:
\verbatim
   void MySend( struct *msg ) \{
      int sclhandle;

      VT_sclstack( NULL, NULL, // we use the default stack unwinding
                   1,          // MySend() is called directly by the
                               // application code we want to trace: 
                               // skip our own source code, but not 
                               // more
                   -1,         // default PCTRACE setting for size 
                               // of recorded stack
                   &sclhandle );
      // if an error occurs, we continue with the sclhandle == VT_NOSCL
      // that VT_sclstack() sets

      VT_enter( funchandle,
                sclhandle );
      VT_log_sendmsg( msg->receiver,
                      msg->count,
                      msg->tag,
                      msg->commid,
                      sclhandle );

      // do the send here

      VT_leave( sclhandle );
   \}
\endverbatim
 *
 * @param pc            record the source code of this program counter value
 *                      as the innermost call location, then continue with
 *                      normal stack unwinding; NULL if only stack unwinding
 *                      is to be used
 * @param stackframe    start unwinding at this stack frame, NULL for starting
 *                      with the stack frame of VT_sclstack() itself:
 *                      on IA32 the stack frame is found in the EBP register,
 *                      on Intel(R) 64 in the RBP register,
 *                      on Itanium(R) in the BSP register
 * @param skip          -1: get the number of stack frames to skip from the PCTRACE
 *                          configuration option
 *                      0: first recorded program counter value after the (optional)
 *                         pc address is the return address of the initial stack frame
 *                      >0: skip the given number of return addresses
 * @param trace         -1: get the number of stack frames to record from the PCTRACE
 *                          configuration option
 *                      0: do not record any source code locations for the call stack:
 *                         returns an SCL ID for the pc address if one is given,
 *                         otherwise returns VT_NOSCL immediately
 *                      >0: the number of stack frames to record
 * @retval sclhandle    the int it points to is set by ITC to a valid
 *                      SCL handle in case of success and VT_NOSCL otherwise
 * @return error code
 */
extern int VT_sclstack( void *pc,
                        void *stackframe,
                        int skip,
                        int trace,
                        int * sclhandle );

/**
 * function to broadcast the new id from given root to other members
 * see shmem_broadcast32 for how to use pestart, log_pestride and pesize
 **/
typedef int (*VT_BcastFunc_t)( int value, int peroot, int pestart, int log_pestride, int pesize );


/**
 * Generate an artificial Communicator.
 * It can be used for API calls loggin messages and collective operations.
 *
 * @param pestart first member (id) in new communicator
 * @praam log_pestride stride between each member (id)
 * @param pesize number of members in communicator
 * @retval commhandle the int points to is set by ITC to a valid
 *                    communicator handle in case of success and VT_COMM_INVALID otherwise
 * @param bcastId function to broadcast the new id 
 * @return error code
 **/
extern int VT_commdef( int pestart, int log_pestride, int pesize, int * commhandle, VT_BcastFunc_t bcastId );


/**
 * Marks the beginning of a region with the name that was assigned to the symbol.
 * Regions should be used to subdivide a function into different parts or
 * to mark the location where a function is called.
 *
 * @par Notes:
 *
 * If automatic tracing of source code locations (aka PC tracing) is
 * supported, then ITC will log the location where VT_begin() is called
 * as source code location for this region and the location where
 * VT_end() is called as SCL for the next part of the calling symbol
 * (which may be a function or another, larger region).
 *
 * If a SCL has been set with VT_thisloc(), then this SCL will
 * be used even if PC tracing is supported.
 *
 * The functions VT_enter() and VT_leave() have been added
 * that can be used to mark the beginning and end of a function call
 * within the function itself. The difference is that a manual source code
 * location which is given to VT_leave() cannot specify where the function
 * call took place, but rather were the function is left. So currently
 * it has to be ignored until the trace file format can store this additional
 * information.
 *
 * If PC tracing is enabled, then the VT_leave routine stores the SCL where
 * the instrumented function was called as SCL for the next part of the
 * calling symbol. In other words, it skips the location where the function
 * is left, which would be recorded if VT_end() were used instead.
 *
 * VT_begin() adds an entry to a stack which can be removed with (and only with)
 * VT_end().
 *
 * @par Fortran
 *  VTBEGIN( statehandle, ierr )
 *
 * @par Java
 *  void begin( int statehandle )
 *
 * @param statehandle handle defined either with VT_symdef() or VT_funcdef()
 * @return error code
 */
extern int  VT_begin( int statehandle );

/**
 * Marks the end of a region.
 *
 * Has to match a VT_begin(). The parameter was used to check this, but this
 * is no longer done to simplify instrumentation; now it is safe to pass a 0
 * instead of the original state handle.
 *
 * @par Fortran
 *  VTEND( statehandle, ierr )
 *
 * @par Java
 *  void end( int statehandle )
 * 
 * @param statehandle    obsolete, pass anything you want
 * @return error code
 */
extern int  VT_end( int statehandle );

/**
 * Mark the beginning of a function. Usage similar to VT_beginl().
 * See also VT_begin().
 *
 * @par Fortran
 *  VTENTER( statehandle, sclhandle, ierr )
 *
 * @par Java
 *  void enter( int statehandle, int sclhandle )
 * 
 * @param statehandle handle defined either with VT_symdef() or VT_funcdef()
 * @param sclhandle handle, defined by VT_scldef.
 *  Use VT_NOSCL if you don't have a specific value.
 * @return error code
 **/
extern int VT_enter( int statehandle, int sclhandle );

/**
 * Mark the end of a function.
 * See also VT_begin().
 *
 * @par Fortran
 *  VTLEAVE( sclhandle, ierr )
 *
 * @par Java
 *  void leave( int sclhandle )
 * 
 * @param sclhandle handle, defined by VT_scldef.
 *  Currently ignored, but is meant to specify
 *  the location of exactly where the function was left
 *  in the future.
 *  Use VT_NOSCL if you don't have a specific value.
 * @return error code
 **/
extern int VT_leave( int sclhandle );

/**
 * Defines a state (when invoked the first time) and enters it.
 * It relies on the caller to provide persistent storage for state
 * handles.
 *
 * The corresponding call to leave the state again is the normal
 * VT_leave(). VT_leave() must be called if and only if VT_enterstate()
 * returns a zero return code.
 *
\verbatim
   static int bsend_handle, bsend_truncated;
   int bsend_retval;
   bsend_retval = VT_enterstate( "MPI:TRANSFER:BSEND",
                                 &bsend_handle, &bsend_truncated );
   ...
   if( !bsend_retval) VT_leave( VT_NOSCL );
\endverbatim
 *
 * As demonstrated in this example, one or more colons (:) may be used to
 * specify parent classes of the state, just as in VT_funcdef() et.al.
 *
 * But in contrast to those, VT_enterstate() also treats a slash (/)
 * as special and uses it to log states at a varying degree of detail:
 * depending on the value of DETAILED-STATES (0 = OFF, 1 = ON, 2,
 * 3...), only the part of the name before the first slash is used
 * (DETAILED-STATES 0). For higher values of DETAILED-STATES more components
 * of the name are used and the slashes in the part of the name which
 * is used is treated like the class separator (:).
 *
 * Examples:
 * - "MPI:TRANSFER/SEND/COPY" + DETAILED-STATES 0: "MPI:TRANSFER"
 * - "MPI:TRANSFER/SEND/COPY" + DETAILED-STATES 1: "MPI:TRANSFER:SEND"
 * - "MPI:TRANSFER/SEND/COPY" + DETAILED-STATES >= 2: "MPI:TRANSFER:SEND:COPY"
 * - "/MPI:INTERNAL" + DETAILED-STATES 0: "" = not logged
 * - "/MPI:INTERNAL" + DETAILED-STATES 1: ":MPI:INTERNAL" = "MPI:INTERNAL"
 *
 * If (and only if) the configuration option DETAILED-STATES causes the
 * truncation of a certain state name, then entering that state is
 * ignored if the process already is in that state.
 *
 * Example of trace with DETAILED-STATES 0:
 * - enter "MPI:TRANSFER/WAIT" = enter "MPI:TRANSFER"
 * - enter "MPI:TRANSFER/COPY" = "MPI:TRANSFER" = ignored by ITC, return code != 0
 * - leave "MPI:TRANSFER/COPY" = ignored by application
 * - enter "MPI:TRANSFER/WAIT" = recursive call; ignored, too
 * - leave "MPI:TRANSFER/WAIT" = ignored by application
 * - leave "MPI:TRANSFER/WAIT" = leave "MPI:TRANSFER"
 *
 * Same trace with DETAILED-STATES 1:
 * - enter "MPI:TRANSFER/WAIT" = enter "MPI:TRANSFER:WAIT"
 * - enter "MPI:TRANSFER/COPY" = enter "MPI:TRANSFER:COPY"
 * - leave "MPI:TRANSFER/COPY" = leave "MPI:TRANSFER:COPY"
 * - enter "MPI:TRANSFER/WAIT" = enter "MPI:TRANSFER:WAIT"
 * - leave "MPI:TRANSFER/WAIT" = leave "MPI:TRANSFER:WAIT"
 * - leave "MPI:TRANSFER/WAIT" = leave "MPI:TRANSFER:WAIT"
 *
 * @par Fortran
 *  VTENTERSTATE( name, statehandle, truncated, ierr )
 *
 * @param name          the name of the state, with colons and/or slashes as
 *                      separators as described above
 * @retval statehandle  must be initialized to zero before calling this
 *                      funtion for the first time, then is set inside the
 *                      function to the state handle which corresponds to 
 *                      the function which is logged
 * @retval truncated    set when calling the function for the first time:
 *                      zero iff the full name is logged
 * @return              zero iff state was entered and VT_leave() needs
 *                      to be called
 */
extern int VT_enterstate( const char *name, int *statehandle, int *truncated );

/**
 * Set source code location for next activity that is logged by ITC.
 * After being logged it is reset to the default behaviour again:
 * automatic PC tracing if enabled in the config file and supported
 * or no SCL otherwise.
 *
 * @par Fortran
 *  VTTHISL( sclhandle, ierr )
 *
 * @par Java
 *  void thisLoc( int sclhandle )
 *
 * @param sclhandle handle defined either with VT_scldef()
 * @return error code
 */
extern int  VT_thisloc( int sclhandle );

/**
 * Shortcut for VT_thisloc( \a sclhandle ); VT_begin( \a statehandle )
 *
 * @par Fortran
 *  VTBEGINL( statehandle, sclhandle, ierr )
 *
 * @par Java
 *  void begin( int statehandle, int sclhandle )
 */
extern int  VT_beginl( int statehandle, int sclhandle );

/**
 * Shortcut for VT_thisloc( \a sclhandle ); VT_end( \a statehandle )
 *
 * @par Fortran
 *  VTENDL( statehandle, sclhandle, ierr )
 *
 * @par Java
 *  void end( int statehandle, int sclhandle )
 */
extern int  VT_endl( int statehandle, int sclhandle );

/** 
 * VT_endl which additionally returns pruning-information.
 *
 * @par Fortran
 *  VTENDP( statehandle, sclhandle, ierr )
 *
 * @par Java
 *  void endP( int statehandle, int sclhandle )
 *
 * @return error code (<=0) or >0 if state was pruned.
 **/
extern int  VT_endP( int statehandle, int sclhandle );

/**
 * Triggers the same additional actions as logging a
 * function call, but without actually logging a call.
 *
 * When ITC logs a function entry or exit it
 * might also execute other actions, like sampling and
 * logging counter data. If a function runs for a very
 * long time, then ITC has no chance to execute
 * these actions. To avoid that, the programmer can
 * insert calls to this function into the source code of
 * the long-running function.
 *
 * @par Fortran
 *  VTWAKEUP( ierr )
 *
 * @par Java
 *  void wakeup()
 *
 * @return error code
 */
extern int VT_wakeup( void );


/**
 * In contrast to a state, which is entered and left with
 * VT_begin/VT_end() resp. VT_enter/VT_leave(), a scope
 * does not follow a stack based approach. It is possible
 * to start a scope "a", then start scope "b" and stop "a"
 * before "b":
\verbatim
    |---- a -----|
      |------ b -----|
\endverbatim
 *
 * A scope is identified by its name and class, just like
 * functions. The source code locations that can be associated
 * with it are just additional and optional attributes; they
 * could be used to mark a static start and end of the scope in
 * the source.
 *
 * As functions, the \a scopename may consist of several components
 * separated by a colon (:).
 *
 * @par Fortran
 *  VTSCOPEDEF( scopename, classhandle, scl1, scl2, scopehandle, ierr )
 *
 * @par Java
 *  int scopeDef( final String scopename, int classhandle,
 *                int scl1, int scl2 )
 *
 * @param scopename     the name of the scope
 * @param classhandle   the class this scope belongs to (defined with VT_classdef())
 * @param scl1          any kind of SCL as defined with VT_scldef(), or VT_NOSCL
 * @param scl2          any kind of SCL as defined with VT_scldef(), or VT_NOSCL
 * @retval scopehandle  set to a numeric handle for the scope, needed by VT_scopebegin()
 * @return error code
 */
extern int VT_scopedef( const char * scopename, int classhandle,
                        int scl1, int scl2, int * scopehandle );

/**
 * Starts a new instance of the scope previously defined with VT_scopedef().
 * There can be more than one instance of a scope at the same time. In
 * order to have the flexibility to stop an arbitrary instance, ITC assigns an
 * intermediate identifier to it which can (but does not have to) be passed
 * to VT_scopeend(). If the application does not need this flexibility, then
 * it can simply pass 0 to VT_scopeend().
 *
 * @par Fortran
 *  VTSCOPEBEGIN( scopehandle, scl, seqnr, ierr )
 *
 * @par Java
 *  int scopeBegin( int scopehandle, int scl )
 *
 * @param scopehandle   the scope as defined by VT_scopedef()
 * @param scl           in contrast to the static SCL given in the scope definition
 *                      this you can vary with each instance; pass VT_NOSCL if not needed
 * @retval seqnr        is set to a number that together with the handle identifies
 *                      the scope instance; pointer may be NULL
 * @return error code
 */
extern int VT_scopebegin( int scopehandle, int scl, int * seqnr );

/**
 * Stops a scope that was previously started with VT_scopebegin().
 *
 * @par Fortran
 *  VTSCOPEEND( scopehandle, seqnr, scl )
 *
 * @par Java
 *  void scopeEnd( int scopehandle, int seqnr, int scl )
 *
 * @param scopehandle   identifies the scope that is to be terminated
 * @param seqnr         0 terminates the most recent scope with the given handle,
 *                      passing the seqnr returned from VT_scopebegin() terminates exactly
 *                      that instance
 * @param scl           a dynamic SCL for leaving the scope
 */
extern int VT_scopeend( int scopehandle, int seqnr, int scl );

/**
 * Returns information about active scopes.
 *
 * Scopes are only remembered if the process has not be turned off
 * completely. If it was turned off completely, then it will always
 * return VT_ERR_BADINDEX even if VT_scopebegin() was used.
 *
 * @sa VT_tracestate()
 *
 * @par Fortran
 *  VTSCOPEQUERY( number, scopehandle, seqnr )
 *
 * @param  number       0 for most recent scope, 1 for next older one...
 *                      Check error code to detect when no older scopes
 *                      are available.
 * @retval scopehandle  set to handle of scope
 * @retval seqnr        set to sequence number of this scope instance
 * @return error code: VT_ERR_BADINDEX if number does not refer to 
 *         existing scope
 */
extern int VT_scopequery( int number, int *scopehandle, int *seqnr );

#if defined(__cplusplus)
} /* extern "C" */

/**
 * C++ utility classes and defines.
 *
 * These are wrappers around the C API calls which simplify
 * instrumentation of C++ source code and ensure correct tracing if
 * exceptions are used.
 */

/**
 * Defines a function on request and then remembers the handle.
 * Can be used to avoid the overhead of defining the function
 * several times in VT_Function.
 */
class VT_FuncDef {
    int m_handle;            /**< stores the function handle, 0 if not defined yet */
    const char *m_symname;   /**< remembers the symbol name until it is needed */
    const char *m_classname; /**< remembers the class name until it is needed */

 public:
    VT_FuncDef( const char *symname, const char *classname ) :
	m_handle(0), m_symname(symname), m_classname(classname)
	{ }

    /**
     * Checks whether the function is defined already or not.
     * Returns handle as soon as it is available, else 0.
     * Defining the function may be impossible e.g. because
     * ITC was not initialized or ran out of memory.
     */
    int GetHandle() {
	int classhandle;
	if( m_handle ||
	    (!VT_classdef( m_classname, &classhandle ) &&
	    !VT_funcdef( m_symname, classhandle, &m_handle ) ) ) {
	    return m_handle;
	} else {
	    return 0;
	}
    }
};

/**
 * Defines a source code location on request and then remembers the handle.
 * Can be used to avoid the overhead of defining the location
 * several times in VT_Function. Best used together with the define
 * VT_SCL_DEF_CXX().
 */
class VT_SclDef {
    int m_handle;            /**< stores the scl handle, 0 if not defined yet */
    const char *m_file;      /**< remembers the file until it is needed */
    int m_line;              /**< remembers the line until it is needed */

 public:
    VT_SclDef( const char *file, int line ) : m_handle(0), m_file(file), m_line(line) { }

    /**
     * Checks whether the scl is defined already or not.
     * Returns handle as soon as it is available, else 0.
     * Defining the function may be impossible e.g. because
     * ITC was not initialized or ran out of memory.
     */
    int GetHandle() {
	if( m_handle ||
	    !VT_scldef( m_file, m_line, &m_handle ) ) {
	    return m_handle;
	} else {
	    return 0;
	}
    }
};

/**
 * This preprocessor macro creates a static source code location definition
 * for the current file and line in C++.
 *
 * @param _sclvar    name of the static variable which is created
 */
#define VT_SCL_DEF_CXX( _sclvar ) static VT_SclDef _sclvar( __FILE__, __LINE__ )

/**
 * In C++ an instance of this class should be created at the beginning
 * of a function. The constructor will then log the function entry,
 * and the destructor the function exit.
 *
 * Providing a source code location for the function exit manually is
 * not supported, because this source code location would have to define
 * where the function returns to. This cannot be determined at compile time.
 */
class VT_Function {
    int m_entered;         /**< entering the function may have failed */

 public:
    /**
     * Defines the function with VT_classdef() and VT_funcdef(),
     * then enters it. This is less efficient than defining the
     * function once and then reusing the handle. Silently ignores
     * errors, like e.g. uninitialized ITC.
     *
     * @param symname     the name of the function
     * @param classname   the class this function belongs to
     */
    VT_Function( const char *symname, const char *classname ) {
	int handle;

	m_entered = 
	    !VT_classdef( classname, &handle ) &&
	    !VT_funcdef( symname, handle, &handle ) &&
	    !VT_enter( handle, VT_NOSCL );
    }

    /**
     * Same as previous constructor, but also stores information about
     * where the function is located in the source code.
     *
     * @param symname     the name of the function
     * @param classname   the class this function belongs to
     * @param file        name of source file, may but does not have to include path
     * @param line        line in this file where function starts
     */
    VT_Function( const char *symname, const char *classname,
		 const char *file, int line ) {
	int handle, scl;
	m_entered =
	    !VT_classdef( classname, &handle ) &&
	    !VT_funcdef( symname, handle, &handle ) &&
	    !VT_scldef( file, line, &scl ) &&
	    !VT_enter( handle, scl );
    }

    /**
     * This is a more efficient version which supports defining the
     * function only once.
     *
     * @param funcdef     this is a reference to the (usually static) instance that
     *                    defines and remembers the function handle
     */
    VT_Function( VT_FuncDef &funcdef ) {
	int handle = funcdef.GetHandle();
	m_entered =
	    handle &&
	    !VT_enter( handle, VT_NOSCL );
    }

    /**
     * This is a more efficient version which supports defining the
     * function and source code location only once.
     *
     * @param funcdef     this is a reference to the (usually static) instance that
     *                    defines and remembers the function handle
     * @param scldef      this is a reference to the (usually static) instance that
     *                    defines and remembers the scl handle
     */
    VT_Function( VT_FuncDef &funcdef, VT_SclDef &scldef ) {
	int funchandle, sclhandle;

	m_entered = 
	    (funchandle = funcdef.GetHandle()) &&
	    (sclhandle = scldef.GetHandle()) &&
	    !VT_enter( funchandle, sclhandle );
    }
    
    /**
     * the destructor marks the function exit
     */
    ~VT_Function( ) {
	if( m_entered ) {
	    VT_leave( VT_NOSCL );
	}
    }
};

/**
 * This is similar to VT_Function, but should be used to mark regions
 * within a function. The difference is that source code locations
 * can be provided for the beginning and end of the region, and one
 * instance of this class can be used to mark several regions in
 * one function.
 */
class VT_Region {
    int m_entered;         /**< entering the region may have failed */

 public:
    /**
     * Defines the region with VT_classdef() and VT_funcdef(),
     * then enters it. This is less efficient than defining the
     * region once and then reusing the handle. Silently ignores
     * errors, like e.g. uninitialized ITC.
     *
     * @param symname     the name of the region
     * @param classname   the class this region belongs to
     */
    void begin( const char *symname, const char *classname ) {
	int handle;

	this->end();
	m_entered = 
	    !VT_classdef( classname, &handle ) &&
	    !VT_funcdef( symname, handle, &handle ) &&
	    !VT_begin( handle );
    }

    /**
     * Same as previous begin(), but also stores information about
     * where the region is located in the source code.
     *
     * @param symname     the name of the region
     * @param classname   the class this region belongs to
     * @param file        name of source file, may but does not have to include path
     * @param line        line in this file where region starts
     */
    void begin( const char *symname, const char *classname,
		const char *file, int line ) {
	int handle, scl;

	this->end();
	m_entered =
	    !VT_classdef( classname, &handle ) &&
	    !VT_funcdef( symname, handle, &handle ) &&
	    !VT_scldef( file, line, &scl ) &&
	    !VT_beginl( handle, scl );
    }

    /**
     * This is a more efficient version which supports defining the
     * region only once.
     *
     * @param funcdef     this is a reference to the (usually static) instance that
     *                    defines and remembers the region handle
     */
    void begin( VT_FuncDef &funcdef ) {
	int handle = funcdef.GetHandle();

	this->end();
	m_entered =
	    handle &&
	    !VT_begin( handle );
    }

    /**
     * This is a more efficient version which supports defining the
     * region and source code location only once.
     *
     * @param funcdef     this is a reference to the (usually static) instance that
     *                    defines and remembers the region handle
     * @param scldef      this is a reference to the (usually static) instance that
     *                    defines and remembers the scl handle
     */
    void begin( VT_FuncDef &funcdef, VT_SclDef &scldef ) {
	int funchandle, sclhandle;

	this->end();
	m_entered = 
	    (funchandle = funcdef.GetHandle()) &&
	    (sclhandle = scldef.GetHandle()) &&
	    !VT_beginl( funchandle, sclhandle );
    }
    

    /**
     * Leaves the region.
     */
    void end( ) {
	if( m_entered ) {
	    VT_end( 0 );
            m_entered = 0;
	}
    }

    /**
     * Same as previous end(), but also stores information about
     * where the region ends in the source code.
     *
     * @param file        name of source file, may but does not have to include path
     * @param line        line in this file where region starts
     */
    void end( const char *file, int line ) {
	int scl;
	if( m_entered &&
	    !VT_scldef( file, line, &scl ) ) {
	    VT_endl( 0, scl );
            m_entered = 0;
	}
    }

    /**
     * This is a more efficient version which supports defining the
     * source code location only once.
     *
     * @param scldef      this is a reference to the (usually static) instance that
     *                    defines and remembers the scl handle
     */
    void end( VT_SclDef &scldef ) {
	int sclhandle;

	if( m_entered &&
	    (sclhandle = scldef.GetHandle()) ) {
	    VT_endl( 0, sclhandle );
            m_entered = 0;
	}
    }

    /**
     * The default constructor does not start the region yet.
     */
    VT_Region( ) : m_entered(0) { }

    /**
     * Enter region when it is created.
     */
    VT_Region( const char *symname, const char *classname ) : m_entered(0)  {
	this->begin( symname, classname );
    }

    /**
     * Same as previous constructor, but also stores information about
     * where the region is located in the source code.
     */
    VT_Region( const char *symname, const char *classname,
		const char *file, int line ) : m_entered(0)  {
	this->begin( symname, classname, file, line );
    }

    /**
     * This is a more efficient version which supports defining the
     * region only once.
     */
    VT_Region( VT_FuncDef &funcdef ) : m_entered(0)  {
	this->begin( funcdef );
    }

    /**
     * This is a more efficient version which supports defining the
     * region and source code location only once.
     */
    VT_Region( VT_FuncDef &funcdef, VT_SclDef &scldef ) : m_entered(0)  {
	this->begin( funcdef, scldef );
    }

    /**
     * the destructor marks the region exit
     */
    ~VT_Region( ) {
	this->end();
    }
};

extern          "C" {
#endif

/**
 * Inserts a enter/leave of "VT_API:..." into the trace file with the current time stamp.
 * This indicates that not all activities within the current state have been logged.
 *
 * @par Fortran
 *  VTPRUNE( ierr )
 *
 * @par Java
 *  void prune()
 *
 * @return error code
 */ 
extern int  VT_prune( void );

/**
 * Logs sending of a message.
 *
 * @par Fortran
 *  VTLOGSENDMSG( other_rank, count, tag, commid, sclhandle, ierr )
 *
 * @par Java
 *  void logSendmsg( int other_rank, int count, int tag, int commid, int sclhandle )
 *
 * @param my_rank rank of the sending process
 * @param other_rank rank of the target process
 * @param count number of bytes sent
 * @param tag tag of the message
 * @param commid numeric ID for the communicator (VT_COMM_WORLD, VT_COMM_SELF, or see VT_commdef())
 * @param sclhandle handle as defined by VT_scldef, or VT_NOSCL
 * @return error code
 */
extern int  VT_log_sendmsg( int other_rank, int count, int tag, int commid, int sclhandle );

/**
 * Logs receiving of a message.
 *
 * @par Fortran
 *  VTLOGRECVMSG( other_rank, count, tag, commid, sclhandle, ierr )
 *
 * @par Java
 *  void logRecvmsg( int other_rank, int count, int tag, int commid, int sclhandle )
 *
 * @param my_rank rank of the receiving process
 * @param other_rank rank of the source process
 * @param count number of bytes sent
 * @param tag tag of the message
 * @param commid numeric ID for the communicator (VT_COMM_WORLD, VT_COMM_SELF, or see VT_commdef())
 * @param sclhandle handle as defined by VT_scldef, or VT_NOSCL
 * @return error code
 */
extern int  VT_log_recvmsg( int other_rank, int count, int tag, int commid, int sclhandle );

/**
 * Insert a comment into the trace file.
 *
 * @par Fortran
 *  VTLOGCOMMENT( comment, ierr )
 *
 * @par Java
 *  void logComment( final String comment )
 *
 * @param comment the comment string to be logged, must not contain newlines
 * @return error code
 **/
extern int  VT_log_comment( const char * comment );

/**
 * Insert opaque data into the trace file.
 * In contrast to the comment string in VT_log_comment(),
 * this data may contain line breaks and 0 bytes.
 *
 * The maximum amount of data that can be stored in one chunk
 * is the same as the configurable
 * buffer size (see config option MEM-BLOCKSIZE) minus less than
 * approx. 64 bytes. At the moment the behaviour when trying to
 * store larger chunks is undefined.
 *
 * The data is found is delivered to the STF_OpaqueDataHandler_t
 * callback.
 *
 * @par Fortran
 *  VTLOGDATA( data, len, sclid, ierr )
 *
 * @par Java
 *  void logData( final byte data[], int sclid )
 *
 * @param data     a pointer to the opaque data
 * @param len      number of bytes to log
 * @param sclid    an source code location id as defined with VT_scldef(), or VT_NOSCL
 * @return error code
 **/
extern int  VT_log_data( const void * data, int len, int sclid  );

/**
 * Starts a period with out-of-order events.
 *
 * Most API functions log events with just one time stamp which is
 * taken when the event is logged. That guarantees strict
 * chronological order of the events.
 *
 * VT_log_msgevent() and VT_log_opevent() are logged when the event
 * has finished with a start time taken earlier with VT_timestamp().
 * This can break the chronological order, e.g. like in the
 * following two examples:
 * \verbatim
   t1: VT_timestamp()     "start message"
   t2: VT_end()              "leave function"
   t3: VT_log_msgevent( t1 ) "finish message"

   t1: VT_timestamp()     "start first message"
   t2: VT_timestamp()     "start second message"
   t3: VT_log_msgevent( t1 ) "finish first message"
   t4: VT_log_msgevent( t2 ) "finish second message"
\endverbatim
 * 
 * In other words, it is okay to just log a complex event if and only
 * if no other event is logged between its start and end in this
 * thread. "logged" in this context includes other complex events
 * that are logged later, but with a start time between the other
 * events start and end time.
 *
 * In all other cases you have to alert ITC of the fact that
 * out-of-order events will follow by calling VT_begin_unordered()
 * before and VT_end_unordered() after these events. When writing the
 * events into the trace file ITC increases a counter per thread when it
 * sees a VT_begin_unordered() and decrease it at a
 * VT_end_unordered(). Events are remembered and sorted until the
 * counter reaches zero, or till the end of the data.
 *
 * This means that:
 * - unordered periods can be nested,
 * - it is not necessary to close each unordered period at
 *   the end of the trace,
 * - but not closing them properly in the middle of a
 *   trace will force ITC to use a lot more memory when
 *   writing the trace (proportional to the number of
 *   events till the end of the trace).
 *
 * @par Fortran
 *  VTBEGINUNORDERED( ierr )
 *
 * @par Java
 *  void beginUnordered()
 */
extern int VT_begin_unordered( void );

/**
 * Close a period with out-of-order events that was started with
 * VT_begin_unordered().
 *
 * @par Fortran
 *  VTENDUNORDERED( ierr )
 *
 * @par Java
 *  void endUnordered()
 */
extern int VT_end_unordered( void );

/**
 * Logs sending and receiving of a message.
 *
 * @par Fortran
 *  VTLOGMSGEVENT( sender, receiver, count, tag, commid, sendts, sendscl, recvscl, ierr )
 *
 * @par Java
 *  void logMsgEvent( int sender, int receiver, int count, int tag, int commid,
 *		      double sendts, int sendscl, int recvscl )
 *
 * @param sender        rank of the sending process
 * @param receiver      rank of the target process
 * @param count         number of bytes sent
 * @param tag           tag of the message
 * @param commid        numeric ID for the communicator (VT_COMM_WORLD, VT_COMM_SELF, or see VT_commdef())
 * @param sendts        time stamp obtained with VT_timestamp() 
 * @param sendscl       handle as defined by VT_scldef() for the source code
 *                      location where the message was sent, or VT_NOSCL
 * @param recvscl       the same for the receive location
 * @return error code
 */
extern int VT_log_msgevent( int sender, int receiver, int count, int tag, int commid,
                            double sendts, int sendscl, int recvscl );

/**
 * Logs the duration and amount of transfered data of an operation for
 * all involved processes at once.
 *
 * ITC knows which processes send and receive data in each
 * operation. Unused byte counts are ignored when writing the trace,
 * so they can be left uninitialized, but NULL is not allowed as
 * array address even if no entry is used at all.
 *
 * @par Fortran
 *  VTLOGOPEVENT( opid, commid, root, numprocs, bsend, brecv, startts, sclhandle, ierr )
 *
 * @par Java
 *  void logOpEvent( int opid, int commid, int root, int numprocs, int bsend[], int brecv[],
 *                   double startts[], int sclhandle )
 *
 * @param opid      id of the operation; must be one of the predefined constants in enum _VT_OpTypes
 * @param commid    numeric ID for the communicator;
 *                  see VT_log_sendmsg() for valid numbers
 * @param root      rank of the root process in the communicator
 *                  (ignored for operations without root, must still be valid, though)
 * @param numprocs  the number of processes in the communicator
 * @param bsend     bytes sent by process
 * @param brecv     bytes received by process
 * @param startts   the start time of the operation (as returned by VT_timestamp())
 * @param sclhandle handle as defined by VT_scldef, or VT_NOSCL
 * @return error code
 */
extern int  VT_log_opevent( int opid, int commid, int root, int numprocs, int *bsend, int *brecv, double *startts, int sclhandle );

/**
 * Logs the duration and amount of transfered data of an operation for
 * one process.
 *
 * @par Fortran
 *  VTLOGOP( opid, commid, root, bsend, brecv, startts, sclhandle, ierr )
 *
 * @par Java
 *  void logOp( int opid, int commid, int root, int bsend, int brecv, double startts, int sclhandle )
 *
 * @param opid      id of the operation; must be one of the predefined constants in enum _VT_OpTypes
 * @param commid    numeric ID for the communicator;
 *                  see VT_log_sendmsg() for valid numbers
 * @param root      rank of the root process in the communicator
 *                  (ignored for operations without root, must still be valid, though)
 * @param bsend     bytes sent by process (ignored for operations that send no data)
 * @param brecv     bytes received by process (ignored for operations that receive no data)
 * @param startts   the start time of the operation (as returned by VT_timestamp())
 * @param sclhandle handle as defined by VT_scldef, or VT_NOSCL
 * @return error code
 */
extern int  VT_log_op( int opid, int commid, int root, int bsend, int brecv, double startts, int sclhandle );

/*@}*/



/**
 * @name API
 */
/*@{*/

/**
 * @name ITC statistics API
 */
/*@{*/

/**
 * Define a new type of statistic. After defining such a stats-type,
 * VT_logstats() can be used to actually store statistics into the 
 * statistics file(s). If you want your statistics to be by process
 * or thread, then you must add these values yourself; ITC will dump
 * your data into the statistic files without annotating them with
 * the process that generated them.
 *
 * @param name           a string to uniquely identify this type of statistics;
 *                       should only use upper case letters and no spaces to match
 *                       the strings used by ITC's own statistics
 * @param num_values     the number of values which are contained in this type of statistics;
 *                       at least one, because the first string entry must be we written
 * @param labels         gives names for the different values (e.g min, max etc.)
 * @param units          units for each values
 * @retval stats_handle  set to a handle for later use with VT_log_stats()
 * @return error code
 **/
extern int VT_statsdef( const char *name, int num_values,
                        const char **labels, const char **units,
                        int *stats_handle );

/**
 * Actually provide statistical data of a certain type.
 *
 * @param stats_handle   specifies the type of statistics (returned by VT_statsdef())
 * @param name           an arbitrary name for this statistic entry
 * @param num_values     number of values in values-array,
 *                       must be the same as provided to VT_statsdef
 * @param values         array of num_values-1 many doubles holding the statistical data;
 *                       the first entry is the name string
 * @return error code
 **/
extern int VT_log_stats( int stats_handle, const char *name,
	                 int num_vals, const double *values );

/**
 * Tells the application whether the user has enabled statistics
 * or not in the initial configuration. The application can ignore
 * this setting, because ITC will write the application's custom
 * statistics even if the user has turned off statistics.
 *
 * @retval on            set to 1 if statistics are enabled (even if
 *                       the process was turned off completely)
 * @return error code
 */
extern int VT_statsstate( int *on );

/*@}*/
/*@}*/

/** @name ITC status API */
/*@{*/

/**
 * Get logging state of current thread.
 * Set by config options PROCESS/CLUSTER,
 * modified by VT_traceon/off().
 *
 * There are three states:
 * - 0 = thread is logging 
 * - 1 = thread is currently not logging 
 * - 2 = logging has been turned off completely
 *
 * Note that different threads within one process 
 * may be in state 0 and 1 at the same time because
 * VT_traceon/off() sets the state of the calling
 * thread, but not for the whole process.
 *
 * State 2 is set via config option "PROCESS/CLUSTER NO"
 * for the whole process and cannot be changed.
 *
 * @par Fortran
 *  VTTRACESTATE( state, ierr )
 *
 * @par Java
 *  int traceState()
 *
 * @retval state is set to current state
 * @return error code
 */
extern int VT_tracestate( int *state );

/**
 * Turn tracing on for thread if it was disabled,
 * otherwise do nothing.
 *
 * Cannot enable tracing if "PROCESS/CLUSTER NO" was
 * applied to the process in the configuration.
 *
 * @par Fortran
 *  VTTRACEON( )
 *
 * @par Java
 *  void traceOn()
 */
extern void VT_traceon( void );

/**
 * Turn tracing off for thread if it was enabled,
 * does nothing otherwise.
 *
 * @par Fortran
 *  VTTRACEOFF( )
 *
 * @par Java
 *  void traceOff()
 */
extern void VT_traceoff( void );

/**
 * Get filter state of one state.
 * Set by config options SYMBOL, ACTIVITY.
 *
 * Note that a state may be active even if
 * the thread's logging state is "off".
 *
 * @par Fortran
 *  VTSYMSTATE( statehandle, on, ierr )
 *
 * @par Java
 *  int symState( int statehandle )
 *
 * @param statehandle result of VT_funcdef() or VT_symdef()
 * @retval on set to 1 if symbol is active
 * @return error code
 */
extern int VT_symstate( int statehandle, int *on );

/*@}*/


/**
 * @name API
 */
/*@{*/

/** @name ITC thread API */
/*@{*/

    /**
     * Registers a new thread with ITC under the given number.
     *
     * Threads are numbered starting from 0, which is always the thread that
     * has called VT_initialize() resp. MPI_Init(). The call to VT_registerthread()
     * is optional: a thread that uses ITC without having called VT_registerthread()
     * is automatically assigned the lowest free index. If a thread terminates,
     * then its index becomes available again and might be reused for another thread.
     *
     * Calling VT_registerthread() when the thread has been assigned an index already
     * is an error, unless the argument of VT_registerthread() is equal to this index.
     * The thread is not (re)registered in case of an error.
     *
     * @par Fortran
     *  VTREGISTERTHREAD( thindex, ierr )
     *
     * @par Java
     *  void registerThread( int thindex )
     *
     * @param thindex      thread number, only used if >= 0
     * @return error code:
     *         - VT_ERR_BADINDEX       - thread index is currently assigned to another thread
     *         - VT_ERR_BADARG         - thread has been assigned a different index already
     *         - VT_ERR_NOTINITIALIZED - ITC wasn't initialized yet
     */
    extern int      VT_registerthread( int thindex );

    /**
     * Registers a new thread with ITC under the given number and name.
     * Threads with the same number cannot have different names. If you try that,
     * the thread uses the number, but not the new name.
     *
     * Trying to register a thread twice with different names or numbers is
     * an error. One can add a name to an already registered thread with
     *    VT_registerthread( "new name", -1 )
     * if no name has been set before.
     *
     * @par Java
     *  void registerThread( final String threadname, int thindex )
     *
     * @param threadname      desired name of the thread, or NULL/empty string if no name wanted
     * @param thindex         desired thread number, pass negative number to let ITC pick a number
     * @return error code, see VT_registerthread()
     */
    extern int      VT_registernamed( const char *threadname, int thindex );
     
    /**
     * Get thread index within process.
     * Either assigned automatically by ITC or manually with VT_registerthread().
     *
     * @par Fortran
     *      VTGETTHRANK( thrank, ierr )
     *
     * @par Java
     *  int getThreadRank()
     *
     * @retval thrank thread index within current thread is stored here
     * @return error code
     */
    extern int VT_getthrank( int *thrank );

    /** obsolete */
    extern int      VT_setthreadnum( int thindex );

/*@}*/
/*@}*/
/**
 * @name API
 */
/*@{*/

/**
 * @name ITC counter API 
 *
 * Generic interface to record time series of values.
 * counters can be grouped into classes (VT_classdef).
 * All values passed to API-functions are assumed to be 64 bit long,
 * either one float or an array of two 4-byte ints. For float values
 * the native byte order is used. For integer both the native
 * 64 bit representation and the representation as pair of 32 bit
 * high and 32 bit low integer (in this order) are supported.
 * The former is simpler to use if the application's programming
 * language has a 64 bit integer type, while the other is easier
 * to store a 32 bit integer: simply define an array of 32 bit integers,
 * set every even (= high) integer to 0 and the odd integers to the
 * real counter values.
 *
 * Counters must be defined using VT_countdef prior to their usage.
 * Upon definition several characteristics of the counter must
 * be declared, such as data type, unit, name etc. It is also
 * possible to request differentiation at display time.
 **/
/*@{*/

enum VT_CountData {
    /**
     * Counter measures 64 bit integer value, passed to ITC API as a pair
     * of high and low 32 bit integers.
     */
    VT_COUNT_INTEGER   = 0x00000000,

    /** Counter measures 64 bit floating point value (native format). */
    VT_COUNT_FLOAT     = 0x00000001,

    /** Counter measures 64 bit integer value (native format). */
    VT_COUNT_INTEGER64 = 0x00000002,

    /** mask to extract the data format */
    VT_COUNT_DATA      = 0x0000000F
};

enum VT_CountDisplay {
    /** counter shall be displayed with absolute values */
    VT_COUNT_ABSVAL  = 0x00000000,

    /** first derivative of counter values shall be displayed */
    VT_COUNT_RATE    = 0x00000010,

    /** mask to extract the display type */
    VT_COUNT_DISPLAY =   0x000000F0
};

enum VT_CountScope {
    /** the value is valid until and at the current time */
    VT_COUNT_VALID_BEFORE = 0x00000000,

    /** the value is valid exactly at the current time,
	and no value is available before or or after it */
    VT_COUNT_VALID_POINT = 0x00000100,

    /** the value is valid at and after the current time */
    VT_COUNT_VALID_AFTER = 0x00000200,

    /** the value is valid at the current time and samples a curve,
	so e.g. linear interpolation between sample values is possible */
    VT_COUNT_VALID_SAMPLE = 0x00000300,

    /** mask to extract the scope */
    VT_COUNT_SCOPE = 0x00000F00
};

/**
 * Define a counter and get handle for it.
 * Counters are identified by their name (string) alone.
 *
 * @par Fortran
 *  VTCOUNTDEF( name, classhandle, genre, target, bounds[], unit, counterhandle, ierr )
 *
 * @param name            string identifying the counter
 * @param classhandle     class to group counters, handle must have been retrieved by VT_classdef
 * @param genre           bitwise or of one value from VT_CountScope, VT_CountDisplay and VT_CountData
 * @param target          target which the counter refers to (VT_ME, VT_GROUP_THREAD, VT_GROUP_PROCESS,
 *                        VT_GROUP_CLUSTER or thread/process-id or user-defined group handle ).
 * @param bounds          array of lower and upper bounds
 *                        (2x 64 bit float, 2x2 32 bit integer, 2x 64 bit integer -> 16 byte)
 * @param unit            string identifying the unit for the counter (like Volt, pints etc.)
 * @retval counterhandle  handle identifying the defined counter
 * @return error code
 **/
extern int VT_countdef( const char * name, int classhandle, int genre,
                        int target, const void * bounds, const char * unit, int * counterhandle );

/**
 * Record counter values.
 * Values are expected as two 4-byte integers, one 8-byte integer or one 8-byte double,
 * according to the counter it refers to.
 *
 * @par Fortran
 *  VTCOUNTVAL( ncounters, handles[], values[], ierr )
 *
 * @param ncounters number of counters to be recorded
 * @param handles array of ncounters many handles (previously defined by VT_countdef)
 * @param values array of ncounters many values, value[i] corresponds to handles[i].
 * @return error code
 **/
extern int VT_countval( int ncounters, int *handles, void *values );

/**
 * Get filter state of one counter (COUNTER).
 *
 * @par Fortran
 *  VTCOUNTSTATE( code, on, ierr )
 *
 * @param code result of VT_countdef()
 * @param on set to 1 if counter is active
 * @return error code
 */
extern int VT_countstate( int code, int *on );

/**
 * Get filter state of one counter (COUNTER), based on the name.
 *
 * @par Fortran
 *  VTCOUNTSTATENAME( name, classhandle, on, ierr )
 *
 * @param name          name of the counter
 * @param classhandle   the class of the counter; it is currently ignored
 *                      and VT_NOCLASS should be passed
 * @param on            set to 1 if counter is active and 0 if not;
 *                      not modified if no config rule applies to it
 * @return error code
 */
extern int VT_countstatename( const char *name, int classhandle, int *on );

/**
 * A callback function that samples counters.
 *
 * Calling VT_countval() from user code is one way to sample
 * counters, but that way counter samples cannot be bound to
 * other events like state changes. The other way is providing
 * ITC with a callback function that ITC can use to sample
 * counters when it needs to.
 *
 * This callback function may sample less, but not more counters than 
 * announced when registering the callback with VT_countsetcallback().
 * It shall put the handles and values for each sampled counter
 * in the following format into memory which is provided by ITC:
 * \verbatim

 +-----------+-----------+-----+-----------+----------+-----+----------+
 | counter 1 | counter 2 | ... | counter n | handle 1 | ... | handle n |
 +-----------+-----------+-----+-----------+----------+-----+----------+
\endverbatim
 * 
 * Each counter value is either a 64 bit float value (VT_COUNT_FLOAT),
 * a pair of high and low 32 bit integers (big endian order,
 * VT_COUNT_INTEGER) or a 64 bit integer (VT_COUNT_INTEGER64).
 * Each handle has 32 bits  which have to be filled with the lower
 * 32 bits of the counter handles returned by VT_countdef(). The \a mem
 * parameter is aligned to a 64 bit boundary to allow direct writing
 * of a 64 bit float or int.
 *
 * ITC will call the function whenever a thread changes its state by
 * calling MPI or entering or leaving a state.
 * The user can enable or disable counter sampling separately for
 * each state in the ITC config file, but cannot
 * control which counters are sampled for a state.
 *
 * ITC logs all samples returned by the callback function, whereas
 * VT_countval() removes those that are disabled by the user.
 *
 * @param custom  the same value as given to VT_countsetcallback()
 * @param ts      current time stamp which will be used in the trace file for the counter values;
 *                format is seconds since process started
 * @param mem     enough memory to store as many counter values and handles as announced
 *                when calling VT_countsetcallback()
 * @param values  array to be filled with counter values (same format as in VT_countval())
 * @return number of counters that were really sampled
 */
typedef unsigned int (* VT_CountCallback_t)( void *custom, double secs, void *mem );

/**
 * Sets a callback for counter sampling for the calling thread.
 *
 * The data provided by the callback is logged with the same
 * time stamp as the event that triggered the callback. The
 * callback must be set for each thread individually. Setting
 * NULL disables sampling for the thread.
 *
 * @param callback  the address of the callback function or NULL
 * @param custom    opaque data that is passed to the callback function by ITC;
 *                  could be used by the callback function to identify the
 *                  thread or enabled counters
 * @param ncounters Upper limit for the number of counters returned by the callback.
 *                  It is not a problem to specify a number that is larger than the
 *                  one that will be actually used, because the extra memory provided to
 *                  the callback function will be reused efficiently.
 * @return error code
 */
extern int VT_countsetcallback( VT_CountCallback_t callback, void *custom, int ncounters );


/*@}*/
/*@}*/


/**
 * @name API
 */
/*@{*/

/** @name ITC grouping API */
/*@{*/

enum VT_Group {
    VT_ME = -4,            /**< the calling thread/process */
    VT_GROUP_THREAD = -1,  /**< Group of all threads */
    VT_GROUP_PROCESS = -2, /**< Group of all processes */
    VT_GROUP_CLUSTER = -3  /**< Group of all clusters */
};

/** 
 * Get global id for process which is identified by process index.
 * If threads are supported, then this id refers to the group of
 * all threads within the process, otherwise the result is identical to
 * VT_getthreadid( procindex, 0, procid ).
 *
 * @par Fortran
 *  VTGETPROCID( procindex, procid, ierr )
 * @par Java
 *  int   getProcID( int procindex );
 * @param  procindex index of process (0 <= procindex < N )
 * @retval procidpointer to mem place where id is written to
 * @return error code
 **/
extern int  VT_getprocid( int procindex, int * procid );

/** 
 * Get global id for the thread which is identified by the pair of process and thread index.
 *
 * @par Fortran
 *  VTGETTHREADID( procindex, thindex, threadid, ierr )
 * @par Java
 *  int getThreadID( int procindex, int thindex )
 *
 * @param  procindex index of process (0 <= procindex < N )
 * @param  thindex   index of thread
 * @retval threadid  pointer to mem place where id is written to
 * @return error code
 **/
extern int  VT_getthreadid( int procindex, int thindex, int * threadid );

/**
 * Defines a new group and returns a handle for it.
 * Groups are distinguished by their name and their members. The order
 * of group members is preserved, which can lead to groups
 * with the same name and same set of members, but different order
 * of these members.
 *
 * @par Fortran
 *  VTGROUPDEF( name, n_members, ids[], grouphandle, ierr )
 *
 * @par Java
 *  int groupDef( final String name, int ids[] )
 *
 * @param  name        the name of the group
 * @param  n_members   number of entries in the \a ids array
 * @param  ids         array where each entry is either:
 * - VT_ME
 * - VT_GROUP_THREAD
 * - VT_GROUP_PROCESS
 * - VT_GROUP_CLUSTER
 * - result of VT_getthreadid(), VT_getprocid() or VT_groupdef()
 * @retval grouphandle handle for the new group, or old handle if the group was defined already
 * @return error code
 */
extern int VT_groupdef( const char *name, int n_members, int * ids, int * grouphandle );

/*@}*/
/*@}*/
/**
 * @name API
 */
/*@{*/

/**
 * @name ITC client/server API
 *
 * These functions are only available in the VTcs library.
 * Each application client has to be linked against this,
 * while the server calls the VTserver application instead
 * of functions.
 */
/*@{*/

/**
 * Initializes communication in a client/server application.
 *
 * Must be called before VT_initialize() in the client of
 * the application. There are three possibilities:
 * -# client is initialized first, which produces a 
 *    contact string that must be passed to the server
 *    (*contact == NULL)
 * -# the server was started first, its contact string is
 *    passed to the clients (*contact == result of VT_serverinit()
 *    with the prefix "S" - this prefix must be added by the application)
 * -# a process spawns children dynamically, its contact string
 *    is given to its children (*contact == result of VT_serverinit()
 *    or VT_clientinit())
 *
 * @param procid     All clients must be enumerated by the application.
 *                   This will become the process id of the local client
 *                   inside its VT_COMM_WORLD. If the VTserver is used, then
 *                   enumeration must start at 1 because VTserver always
 *                   gets rank 0. Threads can be enumerated
 *                   automatically by ITC or by the client by calling
 *                   VT_registerthread().
 * @param clientname The name of the client. Currently only used for
 *                   error messages. Copied by ITC.
 * @retval contact   Will be set to a string which tells other processes 
 *                   how to contact this process. Guaranteed not to contain spaces.
 *                   The client may copy this string, but doesn't have
 *                   to, because ITC will not free this string until
 *                   VT_finalize() is called.
 *
 * @return error code
 */
extern int VT_clientinit( int procid, const char *clientname,
                          const char **contact );

/**
 * Initializes one process as the server that contacts the other
 * processes and coordinates trace file writing. The calling
 * process always gets rank #0.
 *
 * There are two possibilities:
 * -# collect all infos from the clients and pass them here
 *    (numcontacts >= 0, contacts != NULL)
 * -# start the server first, pass its contact string to the clients
 *    (numcontacts >= 0, contacts == NULL)
 *
 * This call replaces starting the VTserver executable in a separate process.
 * Parameters that used to be passed to the VTserver to control tracing
 * and trace writing can be passed to VT_initialize() instead.
 *
 * @param servername     similar to clientname in VT_clientinit():
 *                       the name of the server. Currently only
 *                       used for error messages. Copied by ITC.
 * @param numcontacts    number of client processes
 * @param contacts       contact string for each client process (order is irrelevant);
 *                       copied by ITC
 * @retval contact       Will be set to a string which tells spawned children how to
 *                       contact this server. Guaranteed not to contain spaces.
 *                       The server may copy this string, but doesn't have to, because
 *                       ITC will not free this string until VT_finalize() is called.
 *                       *contact must have been set to NULL before calling this function.
 * @return error code
 */
extern int VT_serverinit( const char *servername, int numcontacts,
                          const char *contacts[], const char **contact );

/**
 * Abort a VT_initialize() or VT_finalize() call running
 * concurrently in a different thread.
 *
 * This call will not block, but it might still take a while before
 * the aborted calls actually return. They will return either
 * successfully (if they have completed without aborting) or
 * with an error code.
 *
 * @return 0 if abort request was sent successfully, else error code
 */
extern int VT_abort( void );

/**
 * Reset ITC no matter what state it is in at the moment,
 * so that VT_clientinit() can be called again.
 *
 * Will block if any other thread is currently
 * executing a ITC call, so take care to abort
 * those first with VT_abort().
 *
 * VT_finalize() will reset ITC automatically.
 */
extern void VT_reset( void );
 
/**
 * Connect to several new processes. These processes
 * must have been spawned already and need to know
 * the contact string of the root process when
 * calling VT_clientinit().
 *
 *    comm == VT_COMM_WORLD is currently not implemented.
 *    It has some design problems: if several children
 *    want to use VT_COMM_WORLD to recursively spawn
 *    more processes, then their parents must also call
 *    VT_attach(), because they are part of this communicator.
 *    If the VTserver is part of the initial VT_COMM_WORLD,
 *    then VT_attach() with VT_COMM_WORLD won't work, because
 *    the VTserver does not know about the spawned processes and
 *    never calls VT_attach().
 *
 * @param root         rank of the process that the spawned
 *                     processes will contact
 * @param comm         either VT_COMM_SELF or VT_COMM_WORLD:
 *                     in the first case root must be 0 and
 *                     the spawned processes are connected
 *                     to just the calling process. In the latter
 *                     case all processes that share this
 *                     VT_COMM_WORLD must call VT_attach()
 *                     and are included in the new communicator.
 *                     root then indicates whose contact infos
 *                     were given to the children.
 * @param numchildren  number of children that the spawning
 *                     processes will wait for
 * @retval childcomm   an identifier for a new communicator
 *                     that includes the parent processes
 *                     in the same order as in their VT_COMM_WORLD,
 *                     followed by the child processes in the order
 *                     specified by their procid argument in
 *                     VT_clientinit(). The spawned processes will
 *                     have access to this communicator via
 *                     VT_get_parent().
 * @return error code
 */
extern int VT_attach( int root, int comm, int numchildren, int *childcomm );

/**
 * Returns the communicator that connects the process with
 * its parent, or VT_COMM_INVALID if not spawned.
 *
 * @retval parentcomm    set to the communicator number that can be used
 *                       to log communication with parents
 * @return error code
 */
extern int VT_get_parent( int *parentcomm );

/*@}*/
/*@}*/

#if defined(__cplusplus)
/* end of C functions */
}
#endif

#endif  /* _VT_H_ */

/* those defines confuse ITC. */
#ifdef shmem_quiet
#undef shmem_quiet
#endif
#ifdef shmem_wait
#undef shmem_wait
#endif

/*@} */



