/***************************************************************************************
 *
 * Intel(R) Trace Collector Library.
 *
 * Copyright(C) 2006-2015 Intel Corporation.
 * All Rights Reserved.
 * 
 * The source code contained or described herein and all documents
 * related to the source code ("Material") are owned by Intel
 * Corporation or its suppliers or licensors. Title to the Material
 * remains with Intel Corporation or its suppliers and licensors. The
 * Material contains trade secrets and proprietary and confidential
 * information of Intel or its suppliers and licensors. The Material
 * is protected by worldwide copyright and trade secret laws and
 * treaty provisions. No part of the Material may be used, copied,
 * reproduced, modified, published, uploaded, posted, transmitted,
 * distributed, or disclosed in any way without Intel's prior express
 * written permission.
 * 
 * No license under any patent, copyright, trade secret or other
 * intellectual property right is granted to or conferred upon you by
 * disclosure or delivery of the Materials, either expressly, by
 * implication, inducement, estoppel or otherwise. Any license under
 * such intellectual property rights must be express and approved by
 * Intel in writing.
 * 
 ***************************************************************************************/

/**
 * @name Replacement for ITC API calls which can (optionally!) be loaded at runtime.
 *
 * The names and parameters of API calls are as documented in VT.h, but when
 * including "VT_dynamic.h" they are replaced with macros which:
 * - call the real implementation from a dynamically loaded ITC library
 *   via a function pointer
 * - evaluate to a 0 (= VT_OK, NULL pointer) if no ITC library has been loaded
 *
 * To use the replacement stub calls:
 * - include VT_dynamic.h instead of VT.h
 * - link with -lVTdynamic -ldl
 * - at runtime call VT_loadlibrary() to load the real ITC library and initialize pointers
 * - to check whether ITC was loaded, call VT_enabled(): it evaluates to true only if an
 *   ITC library has been loaded successfully
 */
/*@{*/

/**
 * @file
 *
 **/
/***************************************************************************************/

#ifndef _VT_DYNAMIC_H_
#define _VT_DYNAMIC_H_

#include <VT.h>

#ifdef __cplusplus
extern "C" {
#endif

#undef VT_loadlibrary

/**
 * Opens the specified ITC library and initializes pointers to all
 * API functions. Fails if library or any API call cannot be found.
 *
 * @param library    name of the shared object to load, may (but does
 *                   not have to) contain path name; example: "libVTcs.so"
 * @return NULL pointer if all went okay, otherwise error description
 */
extern const char *VT_loadlibrary( const char *library );

/*@}*/

struct VT_functions
{
  int ok;
  int (*VT_comm_rank_fp) (int commid, int *rank);
  int (*VT_thisloc_fp) (int sclhandle);
  int (*VT_begin_fp) (int statehandle);
  int (*VT_log_stats_fp) (int stats_handle, const char *name, int num_vals, const double *values);
  int (*VT_commdef_fp) (int pestart, int log_pestride, int pesize, int * commhandle, VT_BcastFunc_t bcastId);
  int (*VT_begin_unordered_fp) (void);
  void (*VT_version_fp) (int appversion, int *vtversion, int *backversion, int *compatible);
  int (*VT_leave_fp) (int sclhandle);
  int (*VT_symdef_fp) (int statehandle, const char * symname, const char * activity);
  const char * (*VT_getdescription_fp) (int type);
  double (*VT_timestamp_fp) (void);
  void (*VT_reset_fp) (void);
  int (*VT_symstate_fp) (int statehandle, int *on);
  int (*VT_wakeup_fp) (void);
  int (*VT_enabled_fp) (void);
  int (*VT_endP_fp) (int statehandle, int sclhandle);
  int (*VT_statsdef_fp) (const char *name, int num_values, const char **labels, const char **units, int *stats_handle);
  int (*VT_get_parent_fp) (int *parentcomm);
  int (*VT_statsstate_fp) (int *on);
  int (*VT_endl_fp) (int statehandle, int sclhandle);
  int (*VT_getprocid_fp) (int procindex, int * procid);
  int (*VT_enterstate_fp) (const char *name, int *statehandle, int *truncated);
  int (*VT_finalize_fp) (void);
  int (*VT_scopebegin_fp) (int scopehandle, int scl, int * seqnr);
  int (*VT_prune_fp) (void);
  int (*VT_countstatename_fp) (const char *name, int classhandle, int *on);
  int (*VT_attach_fp) (int root, int comm, int numchildren, int *childcomm);
  void (*VT_traceon_fp) (void);
  int (*VT_countval_fp) (int ncounters, int *handles, void *values);
  int (*VT_scopequery_fp) (int number, int *scopehandle, int *seqnr);
  int (*VT_serverinit_fp) (const char *servername, int numcontacts, const char *contacts[], const char **contact);
  double (*VT_timestart_fp) (void);
  int (*VT_scopeend_fp) (int scopehandle, int seqnr, int scl);
  void (*VT_traceoff_fp) (void);
  int (*VT_getrank_fp) (int *rank);
  int (*VT_scldef_fp) (const char * file, int line_nr, int * sclhandle);
  int (*VT_classdef_fp) (const char * classname, int * classhandle);
  int (*VT_funcdef_fp) (const char * symname, int classhandle, int * statehandle);
  int (*VT_getthreadid_fp) (int procindex, int thindex, int * threadid);
  int (*VT_clientinit_fp) (int procid, const char *clientname, const char **contact);
  int (*VT_tracestate_fp) (int *state);
  int (*VT_countsetcallback_fp) (VT_CountCallback_t callback, void *custom, int ncounters);
  int (*VT_log_msgevent_fp) (int sender, int receiver, int count, int tag, int commid, double sendts, int sendscl, int recvscl);
  int (*VT_log_comment_fp) (const char * comment);
  int (*VT_comm_size_fp) (int commid, int *size);
  int (*VT_sclstack_fp) (void *pc, void *stackframe, int skip, int trace, int * sclhandle);
  int (*VT_log_sendmsg_fp) (int other_rank, int count, int tag, int commid, int sclhandle);
  int (*VT_log_data_fp) (const void * data, int len, int sclid);
  int (*VT_registernamed_fp) (const char *threadname, int thindex);
  int (*VT_end_fp) (int statehandle);
  int (*VT_end_unordered_fp) (void);
  int (*VT_beginl_fp) (int statehandle, int sclhandle);
  int (*VT_getthrank_fp) (int *thrank);
  int (*VT_timesync_fp) (void);
  int (*VT_groupdef_fp) (const char *name, int n_members, int * ids, int * grouphandle);
  int (*VT_countdef_fp) (const char * name, int classhandle, int genre, int target, const void * bounds, const char * unit, int * counterhandle);
  int (*VT_registerthread_fp) (int thindex);
  int (*VT_log_recvmsg_fp) (int other_rank, int count, int tag, int commid, int sclhandle);
  double (*VT_timeofday_fp) (void);
  int (*VT_flush_fp) (void);
  int (*VT_log_opevent_fp) (int opid, int commid, int root, int numprocs, int *bsend, int *brecv, double *startts, int sclhandle);
  int (*VT_scopedef_fp) (const char * scopename, int classhandle, int scl1, int scl2, int * scopehandle);
  int (*VT_symdefl_fp) (int statehandle, const char * symname, const char * activity, int sclhandle);
  int (*VT_countstate_fp) (int code, int *on);
  int (*VT_initialize_fp) (int * argc, char *** argv);
  int (*VT_abort_fp) (void);
  int (*VT_log_op_fp) (int opid, int commid, int root, int bsend, int brecv, double startts, int sclhandle);
  int (*VT_setthreadnum_fp) (int thindex);
  int (*VT_enter_fp) (int statehandle, int sclhandle);
  int (*VT_setfinalizecallback_fp) (VT_Callback_t callback);
  void * library_handle;
};

extern struct VT_functions __VT_functions;

/* Macro definitions */
#define VT_comm_rank(...) (__VT_functions.ok ? __VT_functions.VT_comm_rank_fp(__VA_ARGS__):(int)0)
#define VT_thisloc(...) (__VT_functions.ok ? __VT_functions.VT_thisloc_fp(__VA_ARGS__):(int)0)
#define VT_begin(...) (__VT_functions.ok ? __VT_functions.VT_begin_fp(__VA_ARGS__):(int)0)
#define VT_log_stats(...) (__VT_functions.ok ? __VT_functions.VT_log_stats_fp(__VA_ARGS__):(int)0)
#define VT_commdef(...) (__VT_functions.ok ? __VT_functions.VT_commdef_fp(__VA_ARGS__):(int)0)
#define VT_begin_unordered(...) (__VT_functions.ok ? __VT_functions.VT_begin_unordered_fp(__VA_ARGS__):(int)0)
#define VT_version(...) (__VT_functions.ok ? __VT_functions.VT_version_fp(__VA_ARGS__):(void)0)
#define VT_leave(...) (__VT_functions.ok ? __VT_functions.VT_leave_fp(__VA_ARGS__):(int)0)
#define VT_symdef(...) (__VT_functions.ok ? __VT_functions.VT_symdef_fp(__VA_ARGS__):(int)0)
#define VT_getdescription(...) (__VT_functions.ok ? __VT_functions.VT_getdescription_fp(__VA_ARGS__):(const char *)0)
#define VT_timestamp(...) (__VT_functions.ok ? __VT_functions.VT_timestamp_fp(__VA_ARGS__):(double)0)
#define VT_reset(...) (__VT_functions.ok ? __VT_functions.VT_reset_fp(__VA_ARGS__):(void)0)
#define VT_symstate(...) (__VT_functions.ok ? __VT_functions.VT_symstate_fp(__VA_ARGS__):(int)0)
#define VT_wakeup(...) (__VT_functions.ok ? __VT_functions.VT_wakeup_fp(__VA_ARGS__):(int)0)
#define VT_enabled(...) (__VT_functions.ok ? __VT_functions.VT_enabled_fp(__VA_ARGS__):(int)0)
#define VT_endP(...) (__VT_functions.ok ? __VT_functions.VT_endP_fp(__VA_ARGS__):(int)0)
#define VT_statsdef(...) (__VT_functions.ok ? __VT_functions.VT_statsdef_fp(__VA_ARGS__):(int)0)
#define VT_get_parent(...) (__VT_functions.ok ? __VT_functions.VT_get_parent_fp(__VA_ARGS__):(int)0)
#define VT_statsstate(...) (__VT_functions.ok ? __VT_functions.VT_statsstate_fp(__VA_ARGS__):(int)0)
#define VT_endl(...) (__VT_functions.ok ? __VT_functions.VT_endl_fp(__VA_ARGS__):(int)0)
#define VT_getprocid(...) (__VT_functions.ok ? __VT_functions.VT_getprocid_fp(__VA_ARGS__):(int)0)
#define VT_enterstate(...) (__VT_functions.ok ? __VT_functions.VT_enterstate_fp(__VA_ARGS__):(int)0)
#define VT_finalize(...) (__VT_functions.ok ? __VT_functions.VT_finalize_fp(__VA_ARGS__):(int)0)
#define VT_scopebegin(...) (__VT_functions.ok ? __VT_functions.VT_scopebegin_fp(__VA_ARGS__):(int)0)
#define VT_prune(...) (__VT_functions.ok ? __VT_functions.VT_prune_fp(__VA_ARGS__):(int)0)
#define VT_countstatename(...) (__VT_functions.ok ? __VT_functions.VT_countstatename_fp(__VA_ARGS__):(int)0)
#define VT_attach(...) (__VT_functions.ok ? __VT_functions.VT_attach_fp(__VA_ARGS__):(int)0)
#define VT_traceon(...) (__VT_functions.ok ? __VT_functions.VT_traceon_fp(__VA_ARGS__):(void)0)
#define VT_countval(...) (__VT_functions.ok ? __VT_functions.VT_countval_fp(__VA_ARGS__):(int)0)
#define VT_scopequery(...) (__VT_functions.ok ? __VT_functions.VT_scopequery_fp(__VA_ARGS__):(int)0)
#define VT_serverinit(...) (__VT_functions.ok ? __VT_functions.VT_serverinit_fp(__VA_ARGS__):(int)0)
#define VT_timestart(...) (__VT_functions.ok ? __VT_functions.VT_timestart_fp(__VA_ARGS__):(double)0)
#define VT_scopeend(...) (__VT_functions.ok ? __VT_functions.VT_scopeend_fp(__VA_ARGS__):(int)0)
#define VT_traceoff(...) (__VT_functions.ok ? __VT_functions.VT_traceoff_fp(__VA_ARGS__):(void)0)
#define VT_getrank(...) (__VT_functions.ok ? __VT_functions.VT_getrank_fp(__VA_ARGS__):(int)0)
#define VT_scldef(...) (__VT_functions.ok ? __VT_functions.VT_scldef_fp(__VA_ARGS__):(int)0)
#define VT_classdef(...) (__VT_functions.ok ? __VT_functions.VT_classdef_fp(__VA_ARGS__):(int)0)
#define VT_funcdef(...) (__VT_functions.ok ? __VT_functions.VT_funcdef_fp(__VA_ARGS__):(int)0)
#define VT_getthreadid(...) (__VT_functions.ok ? __VT_functions.VT_getthreadid_fp(__VA_ARGS__):(int)0)
#define VT_clientinit(...) (__VT_functions.ok ? __VT_functions.VT_clientinit_fp(__VA_ARGS__):(int)0)
#define VT_tracestate(...) (__VT_functions.ok ? __VT_functions.VT_tracestate_fp(__VA_ARGS__):(int)0)
#define VT_countsetcallback(...) (__VT_functions.ok ? __VT_functions.VT_countsetcallback_fp(__VA_ARGS__):(int)0)
#define VT_log_msgevent(...) (__VT_functions.ok ? __VT_functions.VT_log_msgevent_fp(__VA_ARGS__):(int)0)
#define VT_log_comment(...) (__VT_functions.ok ? __VT_functions.VT_log_comment_fp(__VA_ARGS__):(int)0)
#define VT_comm_size(...) (__VT_functions.ok ? __VT_functions.VT_comm_size_fp(__VA_ARGS__):(int)0)
#define VT_sclstack(...) (__VT_functions.ok ? __VT_functions.VT_sclstack_fp(__VA_ARGS__):(int)0)
#define VT_log_sendmsg(...) (__VT_functions.ok ? __VT_functions.VT_log_sendmsg_fp(__VA_ARGS__):(int)0)
#define VT_log_data(...) (__VT_functions.ok ? __VT_functions.VT_log_data_fp(__VA_ARGS__):(int)0)
#define VT_registernamed(...) (__VT_functions.ok ? __VT_functions.VT_registernamed_fp(__VA_ARGS__):(int)0)
#define VT_end(...) (__VT_functions.ok ? __VT_functions.VT_end_fp(__VA_ARGS__):(int)0)
#define VT_end_unordered(...) (__VT_functions.ok ? __VT_functions.VT_end_unordered_fp(__VA_ARGS__):(int)0)
#define VT_beginl(...) (__VT_functions.ok ? __VT_functions.VT_beginl_fp(__VA_ARGS__):(int)0)
#define VT_getthrank(...) (__VT_functions.ok ? __VT_functions.VT_getthrank_fp(__VA_ARGS__):(int)0)
#define VT_timesync(...) (__VT_functions.ok ? __VT_functions.VT_timesync_fp(__VA_ARGS__):(int)0)
#define VT_groupdef(...) (__VT_functions.ok ? __VT_functions.VT_groupdef_fp(__VA_ARGS__):(int)0)
#define VT_countdef(...) (__VT_functions.ok ? __VT_functions.VT_countdef_fp(__VA_ARGS__):(int)0)
#define VT_registerthread(...) (__VT_functions.ok ? __VT_functions.VT_registerthread_fp(__VA_ARGS__):(int)0)
#define VT_log_recvmsg(...) (__VT_functions.ok ? __VT_functions.VT_log_recvmsg_fp(__VA_ARGS__):(int)0)
#define VT_timeofday(...) (__VT_functions.ok ? __VT_functions.VT_timeofday_fp(__VA_ARGS__):(double)0)
#define VT_flush(...) (__VT_functions.ok ? __VT_functions.VT_flush_fp(__VA_ARGS__):(int)0)
#define VT_log_opevent(...) (__VT_functions.ok ? __VT_functions.VT_log_opevent_fp(__VA_ARGS__):(int)0)
#define VT_scopedef(...) (__VT_functions.ok ? __VT_functions.VT_scopedef_fp(__VA_ARGS__):(int)0)
#define VT_symdefl(...) (__VT_functions.ok ? __VT_functions.VT_symdefl_fp(__VA_ARGS__):(int)0)
#define VT_countstate(...) (__VT_functions.ok ? __VT_functions.VT_countstate_fp(__VA_ARGS__):(int)0)
#define VT_initialize(...) (__VT_functions.ok ? __VT_functions.VT_initialize_fp(__VA_ARGS__):(int)0)
#define VT_abort(...) (__VT_functions.ok ? __VT_functions.VT_abort_fp(__VA_ARGS__):(int)0)
#define VT_log_op(...) (__VT_functions.ok ? __VT_functions.VT_log_op_fp(__VA_ARGS__):(int)0)
#define VT_setthreadnum(...) (__VT_functions.ok ? __VT_functions.VT_setthreadnum_fp(__VA_ARGS__):(int)0)
#define VT_enter(...) (__VT_functions.ok ? __VT_functions.VT_enter_fp(__VA_ARGS__):(int)0)
#define VT_setfinalizecallback(...) (__VT_functions.ok ? __VT_functions.VT_setfinalizecallback_fp(__VA_ARGS__):(int)0)

#undef VT_enabled
#define VT_enabled() (__VT_functions.ok)


#ifdef __cplusplus
}
#endif

#endif /* _VT_DYNAMIC_H_ */

